package at.ac.uibk.gitsearch.config.prometheus;

import io.micrometer.core.instrument.Clock;
import io.micrometer.core.instrument.Tag;
import io.micrometer.core.instrument.config.MeterFilter;
import io.micrometer.prometheus.PrometheusConfig;
import io.micrometer.prometheus.PrometheusMeterRegistry;
import io.prometheus.client.CollectorRegistry;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.actuate.metrics.export.prometheus.PrometheusScrapeEndpoint;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;

/**
 * Configures Prometheus to use more than the default endpoint
 * to scrape metrics with different time intervals.
 *
 * @author Johannes Kainz
 */
@Configuration
public class PrometheusConfiguration {

    private static final String TAG_INTERVAL = "interval";
    private static final String TAG_HOURLY_INTERVAL = "hourly";

    @SuppressWarnings("unused")
    private final Logger log = LoggerFactory.getLogger(PrometheusConfiguration.class);

    /**
     * represents all metrics which should be scraped from the default prometheus endpoint
     */
    private final PrometheusMeterRegistry myDefaultMeterRegistry;
    /**
     * represents all metrics which should be scraped form the custom hourly scrape endpoint
     */
    private final PrometheusMeterRegistry myHourlyMeterRegistry;

    public PrometheusConfiguration(PrometheusConfig config, Clock clock) {
        this.myDefaultMeterRegistry = new PrometheusMeterRegistry(config, new CollectorRegistry(true), clock);
        this.myDefaultMeterRegistry.config()
            .meterFilter(MeterFilter.denyUnless(id -> id.getTags().stream().noneMatch(tag -> TAG_INTERVAL.equals(tag.getKey()))));

        this.myHourlyMeterRegistry = new PrometheusMeterRegistry(config, new CollectorRegistry(true), clock);
        this.myHourlyMeterRegistry.config()
            .meterFilter(MeterFilter.denyUnless(id -> id.getTags().contains(Tag.of(TAG_INTERVAL, TAG_HOURLY_INTERVAL))));
    }

    @Bean
    PrometheusMeterRegistry defaultMeterRegistry() {
        return this.myDefaultMeterRegistry;
    }

    @Bean
    PrometheusMeterRegistry hourlyMeterRegistry() {
        return this.myHourlyMeterRegistry;
    }

    @Bean
    PrometheusScrapeEndpoint defaultPrometheusScrapeEndpoint() {
        return new PrometheusScrapeEndpoint(this.myDefaultMeterRegistry.getPrometheusRegistry());
    }

    @Bean
    CustomPrometheusScrapeEndpoint hourlyPrometheusScrapeEndpoint() {
        return new HourlyPrometheusScrapeEndpoint(this.myHourlyMeterRegistry.getPrometheusRegistry());
    }
}
