package at.ac.uibk.gitsearch.service;

import co.elastic.clients.elasticsearch._types.ElasticsearchException;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.Metrics;
import io.micrometer.core.instrument.Tag;
import io.micrometer.core.instrument.Tags;
import java.io.IOException;
import java.net.ConnectException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Stream;
import javax.annotation.PostConstruct;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

/**
 * Service class for monitoring
 *
 * @author Johannes Kainz
 */
@Service
// @Transactional // enabling this may cause NullPointerException, because of wrongly applied EnhancerBySpringCGLIB
public class MonitoringService {

    private static final String UNCLASSIFIED = "Unclassified";
    private static final String TAG_INTERVAL = "interval";
    private static final String TAG_HOURLY_INTERVAL = "hourly";

    public static final String USER_TAG = "classification";
    public static final String LANGUAGE_TAG = "language";

    public static final String NUMBER_USER_NAME = "number.users";
    public static final String NUMBER_EXERCISE_NAME = "number.exercises";
    public static final String NUMBER_LIKE_NAME = "number.likes";
    public static final String NUMBER_VIEW_NAME = "number.views";
    public static final String NUMBER_DOWNLOAD_NAME = "number.downloads";

    public static final Tag HOURLY_TAG = Tag.of(TAG_INTERVAL, TAG_HOURLY_INTERVAL);

    private static final Logger LOGGER = LoggerFactory.getLogger(MonitoringService.class);

    private final UserService userService;
    private final SearchService searchService;
    private final LikesService likesService;
    private final StatisticsService statisticsService;

    private final MeterRegistry meterRegistry;

    // the following elements must be stored explizit, to make sure
    // that they are not garbage collected
    private final AtomicInteger allUsers = new AtomicInteger(0);
    private final AtomicInteger activeUsers = new AtomicInteger(0);
    private final AtomicInteger allLikes = new AtomicInteger(0);
    private final AtomicInteger allViews = new AtomicInteger(0);
    private final AtomicInteger allDownloads = new AtomicInteger(0);
    private final AtomicInteger allExercises = new AtomicInteger(0);
    private final Map<String, AtomicInteger> allLanguageCount = new ConcurrentHashMap<>();

    @SuppressWarnings("PMD.AvoidDuplicateLiterals")
    public MonitoringService(
        MeterRegistry meterRegistry,
        UserService userService,
        SearchService searchService,
        LikesService likesService,
        StatisticsService statisticsService
    ) {
        this.meterRegistry = meterRegistry;
        this.userService = userService;
        this.searchService = searchService;
        this.likesService = likesService;
        this.statisticsService = statisticsService;
    }

    @PostConstruct
    public void initMonitoring() {
        initializeGauges();

        monitorHourlyMetrics();
    }

    private void initializeGauges() {
        meterRegistry.gauge(NUMBER_USER_NAME, Tags.of(HOURLY_TAG, Tag.of(USER_TAG, "All")), allUsers);
        meterRegistry.gauge(NUMBER_USER_NAME, Tags.of(HOURLY_TAG, Tag.of(USER_TAG, "Active")), activeUsers);
        meterRegistry.gauge(NUMBER_LIKE_NAME, Tags.of(HOURLY_TAG), allLikes);
        meterRegistry.gauge(NUMBER_VIEW_NAME, Tags.of(HOURLY_TAG), allViews);
        meterRegistry.gauge(NUMBER_DOWNLOAD_NAME, Tags.of(HOURLY_TAG), allDownloads);
        meterRegistry.gauge(NUMBER_EXERCISE_NAME, Tags.of(HOURLY_TAG, Tag.of(LANGUAGE_TAG, "All")), allExercises);
    }

    /**
     * Monitors metrics with an hourly cron job
     */
    @Scheduled(cron = "0 30 * * * ?")
    @SuppressWarnings({ "PMD.AvoidCatchingGenericException", "PMD.CyclomaticComplexity" })
    public final void monitorHourlyMetrics() {
        LOGGER.debug("Started monitoring cron job");

        allUsers.set((int) this.userService.countAllUsers());
        activeUsers.set((int) this.userService.countAllActiveUsers());
        allLikes.set((int) likesService.countAll());
        final List<Integer[]> numberOfDownloadsAndViews = statisticsService.findNumberOfDownloadsAndViews();
        if (numberOfDownloadsAndViews != null && numberOfDownloadsAndViews.size() == 1 && numberOfDownloadsAndViews.get(0).length == 2) {
            if (numberOfDownloadsAndViews.get(0)[1] != null) {
                allViews.set(numberOfDownloadsAndViews.get(0)[1]);
            }

            if (numberOfDownloadsAndViews.get(0)[0] != null) {
                allDownloads.set(numberOfDownloadsAndViews.get(0)[0]);
            }
        }

        AtomicInteger exerciseCounter = new AtomicInteger(0);
        try {
            Map<String, Integer> languageCount = new ConcurrentHashMap<>();
            try {
                final Stream<SearchResultDTO> searchResults = this.getAllSearchResults();

                searchResults.forEach(exercise -> {
                    exerciseCounter.addAndGet(1);
                    updateLanguageCount(exercise, languageCount);
                });
                this.setExercisesLanguageCount(languageCount);
            } catch (ConnectException e) {
                LOGGER.warn("An Connection Exception occurred {}", e.getMessage());
            }
        } catch (IOException | ElasticsearchException exception) {
            LOGGER.warn("An IOException occurred", exception);
            return;
        } catch (org.elasticsearch.ElasticsearchException exception) {
            // May become deprecated soon
            LOGGER.warn("An org.elasticsearch.ElasticsearchException occurred", exception);
            return;
        } catch (Exception exception) {
            LOGGER.warn("An Exception occurred", exception);
            return;
        }

        allExercises.set(exerciseCounter.get());

        LOGGER.debug("Finished monitoring cron job");
    }

    private void updateLanguageCount(SearchResultDTO exercise, Map<String, Integer> languageCount) {
        var metaData = exercise.getMetadata();
        String[] languageMetaData = metaData.getProgrammingLanguage();
        if (languageMetaData == null) {
            languageCount.put(UNCLASSIFIED, languageCount.getOrDefault(UNCLASSIFIED, 0) + 1);
        } else {
            Arrays
                .stream(metaData.getProgrammingLanguage())
                .forEach(language -> languageCount.put(language, languageCount.getOrDefault(language, 0) + 1));
        }
    }

    /**
     * Sets the per programming language exercise counters for given search results
     *
     * @param searchResults the search results
     */
    private void setExercisesLanguageCount(Map<String, Integer> languageCount) {
        synchronized (this) {
            languageCount
                .entrySet()
                .stream()
                .forEach(entry -> {
                    final AtomicInteger existingEntry = allLanguageCount.get(entry.getKey());
                    if (existingEntry == null) {
                        final AtomicInteger newEntry = new AtomicInteger(entry.getValue());
                        Metrics.gauge(NUMBER_EXERCISE_NAME, Tags.of(HOURLY_TAG, Tag.of(LANGUAGE_TAG, entry.getKey())), newEntry);
                        allLanguageCount.put(entry.getKey(), newEntry);
                    } else {
                        existingEntry.set(entry.getValue());
                    }
                });
            // reset outdated languages to zero
            allLanguageCount
                .entrySet()
                .stream()
                .forEach(entry -> {
                    if (!languageCount.containsKey(entry.getKey())) {
                        entry.getValue().set(0);
                    }
                });
        }
    }

    /**
     * Gets the search results of a search for all
     *
     * @return the search results DTO
     * @throws IOException exception for timeout etc.
     */
    private Stream<SearchResultDTO> getAllSearchResults() throws IOException {
        return this.searchService.getAllResultsAsStream();
    }
}
