package at.ac.uibk.gitsearch.repository.search.testESService;

import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.request;

import at.ac.uibk.gitsearch.domain.util.TreeNode;
import at.ac.uibk.gitsearch.repository.search.ElasticSearchRepository;
import at.ac.uibk.gitsearch.repository.search.MetaDataRepository;
import at.ac.uibk.gitsearch.repository.search.SearchRepositoryConstants;
import co.elastic.clients.elasticsearch.ElasticsearchClient;
import co.elastic.clients.elasticsearch._types.query_dsl.BoolQuery;
import co.elastic.clients.elasticsearch.core.SearchResponse;
import co.elastic.clients.transport.rest_client.RestClientTransport;
import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import com.github.dockerjava.api.command.CreateContainerCmd;
import com.github.dockerjava.api.model.ExposedPort;
import com.github.dockerjava.api.model.PortBinding;
import com.github.dockerjava.api.model.Ports;
import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.net.ServerSocket;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.HttpHost;
import org.codeability.sharing.plugins.api.search.GitProjectDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO.MetadataFile;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.codeability.sharing.plugins.api.search.util.ExerciseId;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.node.InternalSettingsPreparer;
import org.elasticsearch.node.Node;
import org.elasticsearch.node.NodeValidationException;
import org.elasticsearch.plugins.Plugin;
import org.gitlab4j.api.models.Visibility;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.context.event.ApplicationStartingEvent;
import org.springframework.context.event.ContextStartedEvent;
import org.springframework.context.event.ContextStoppedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;
import org.springframework.util.Assert;
import org.springframework.web.client.RestTemplate;
import org.testcontainers.containers.wait.strategy.HttpWaitStrategy;
import org.testcontainers.containers.wait.strategy.WaitStrategy;
import org.testcontainers.elasticsearch.ElasticsearchContainer;
import org.testcontainers.utility.DockerImageName;
import org.testcontainers.utility.TestcontainersConfiguration;

/**
 * this class provides test infrastructure for a local elastic search service
 * for local tests.
 *
 * @author Michael Breu
 *
 */
@Service
public class ElasticSearchTestConfiguration {

    static final DockerImageName ELASTICSEARCH_IMAGE = DockerImageName.parse("docker.elastic.co/elasticsearch/elasticsearch:8.8.1");

    private ElasticsearchContainer esContainer;

    @Autowired
    private ElasticsearchClient elasticsearchAPIClient;

    @Autowired
    private ElasticSearchRepository elasticSearchRepository;

    @Autowired
    MetaDataRepository metaDataRepository;

    @Value("${spring.elasticsearch.rest.uris}")
    private String elasticSearchTestURI;

    public void stopTestContainer() {
        if (esContainer != null && esContainer.isCreated() && esContainer.isRunning()) {
            try {
                BoolQuery.Builder b = new BoolQuery.Builder();
                SearchResponse<SearchResultDTO> searchResponse = elasticsearchAPIClient.search(
                    search -> search.index(SearchRepositoryConstants.INDEX_METADATA).query(q -> q.bool(b.build())).from(0).size(1000),
                    SearchResultDTO.class
                );
                @SuppressWarnings("null")
                long totalHits = searchResponse.hits().total().value();
                LOGGER.info("Test ES still containing {} entries before shut down", totalHits);
            } catch (IOException e1) {
                LOGGER.info("Last Search not working", e1);
            }

            esContainer.stop();
            esContainer.close();
            while (esContainer.isRunning()) {
                try {
                    Thread.sleep(1000);
                    LOGGER.info("Shutting down esContainer ...");
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
            esContainer = null;
            LOGGER.warn("Shutted down ES Test Node");
        }
    }

    private static int containerCounter = 0;

    private int getElasticSearchPort() {
        try {
            URI es = URI.create(elasticSearchTestURI.split(",")[0]);
            return es.getPort();
        } catch (IllegalArgumentException | NullPointerException e) {
            LOGGER.warn("Cannot parse " + elasticSearchTestURI + " to get the port");
            return 9200;
        }
    }

    public boolean isAvailable(int portNr) {
        boolean portFree;
        try (var ignored = new ServerSocket(portNr)) {
            portFree = true;
        } catch (IOException e) {
            portFree = false;
        }
        return portFree;
    }

    @SuppressWarnings("resource")
    protected HttpHost startTestESContainer() throws IOException {
        TestcontainersConfiguration.getInstance().getClientPingTimeout();
        // TestcontainersConfiguration.getInstance().updateUserConfig(getMetaDataConfigDefinition(),
        // getMetaDataConfigDefinition());

        long memSize = 2 * 1024 * 1024 * 1024L;

        Consumer<CreateContainerCmd> memoryCmd = e -> e.getHostConfig().withMemory(memSize);

        WaitStrategy waitStrategy = new HttpWaitStrategy()
            .withReadTimeout(Duration.of(10, ChronoUnit.SECONDS))
            .withStartupTimeout(Duration.of(5, ChronoUnit.MINUTES));

        final int hostPort = getElasticSearchPort();
        int containerExposedPort = 9200;
        Consumer<CreateContainerCmd> cmd = e ->
            e.getHostConfig().withPortBindings(new PortBinding(Ports.Binding.bindPort(hostPort), new ExposedPort(containerExposedPort)));

        esContainer =
            new ElasticsearchContainer(ELASTICSEARCH_IMAGE)
                .withStartupAttempts(5)
                .withExposedPorts(9200)
                .withCreateContainerCmdModifier(cmd)
                .withCreateContainerCmdModifier(memoryCmd)
                .withImagePullPolicy(name -> {
                    return false;
                })
                .waitingFor(waitStrategy)
                .withEnv("xpack.security.enabled", "false")
                .withCreateContainerCmdModifier(e -> e.withName("elasticsearchTestContainer" + (containerCounter++)));

        // Start the container. This step might take some time...
        esContainer.start();

        //wait for container to run
        while (!esContainer.isRunning()) {
            try {
                Thread.sleep(1000);
                LOGGER.info("Starting test container...");
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }

        Integer esPort = esContainer.getMappedPort(9200);

        String ipAddress = esContainer.getContainerIpAddress();

        LOGGER.info("ES port is {}@{}", esPort, ipAddress);

        return new HttpHost(ipAddress, esPort);
    }

    private static final org.slf4j.Logger LOGGER = LoggerFactory.getLogger(ElasticSearchTestConfiguration.class);
    private TestESNode testNode;

    public ElasticSearchTestConfiguration() {}

    public synchronized void startTestNode() throws IOException, NodeValidationException {
        try {
            if (!isAvailable(29200)) {
                LOGGER.warn("Port 29200 is not available. Trying to stop test container.");
                stopTestContainer();
            }
            HttpHost hostURL = startTestESContainer();

            setUpESClientForTestESServer(hostURL);

            setUpMetaDataIndex(hostURL);
            setUpContent(hostURL);
            metaDataRepository.updateAutocompletionCache();
            LOGGER.info("Started ES Test Node");
        } catch (Throwable t) {
            LOGGER.info("ES Test Node crashed: ", t);
            stopTestContainer();
        }
    }

    /**
     * redirects the clients to the docker instance.
     * Is there are more elegant way to mock es node urls?
     *
     * @param hostURL
     * @throws IOException
     */
    private void setUpESClientForTestESServer(HttpHost hostURL) throws IOException {
        org.elasticsearch.client.Node n = new org.elasticsearch.client.Node(new HttpHost(hostURL));

        final RestClientTransport esTransport = (RestClientTransport) elasticsearchAPIClient._transport();
        esTransport.restClient().setNodes(Collections.singleton(n));
    }

    private void setUpMetaDataIndex(HttpHost hostURL) throws IOException {
        String indexName = elasticSearchRepository.createNewMetaDataIndex();
        elasticSearchRepository.createAlias(indexName);
    }

    int contentCount = 0;

    private void setUpContent(HttpHost elURL) throws IOException, URISyntaxException {
        URL test = this.getClass().getResource("../../../service/testData");
        File rootData = new File(test.toURI());
        File metaDataTestDir = new File(rootData, "metaData");

        List<SearchResultDTO> contentToIndex = new ArrayList<>();

        contentCount = 0;
        visitSubdirectories(
            metaDataTestDir,
            (File metaDataFile, List<TreeNode<SearchResultDTO>> descendants, String relativePath, GitProjectDTO gitProject) -> {
                try {
                    final UserProvidedMetadataDTO userMetaData = parseMetaDataFile(metaDataFile);

                    SearchResultDTO toIndex = new SearchResultDTO();
                    toIndex.setMetadata(userMetaData);

                    final ExerciseId exerciseId = calculateProjectId(relativePath);

                    if (gitProject == null) {
                        gitProject = new GitProjectDTO();

                        gitProject.setProject_id(Integer.parseInt(exerciseId.getProjectId()));
                        gitProject.setProject_name("SomeGitProject" + exerciseId.getProjectId());
                        gitProject.setVisibility(Visibility.PUBLIC);
                    }

                    toIndex.setProject(gitProject);

                    toIndex.setExerciseId(exerciseId.toString());

                    final MetadataFile file = new MetadataFile();
                    file.setFilename(metaDataFile.getName());
                    // path currently starts with project id
                    int slashPos = relativePath.indexOf("/");
                    String projectRelativePath = (slashPos < 0)
                        ? metaDataFile.getName()
                        : (relativePath.substring(slashPos + 1) + "/" + metaDataFile.getName());
                    file.setPath(projectRelativePath);
                    file.setIndexing_date(Instant.now());
                    file.setCommit_id(gitProject.getCommit_id());
                    file.setChildren(
                        descendants
                            .stream()
                            .map(node -> node.getData().getExerciseId())
                            .collect(Collectors.toList())
                            .toArray(new String[] {})
                    );
                    descendants.forEach(node -> node.getData().getFile().setParentId(exerciseId.toString()));
                    toIndex.setFile(file);

                    toIndex.setExerciseId(exerciseId.toString());
                    contentToIndex.add(toIndex);
                    ElasticSearchTestConfiguration.this.contentCount++;
                    return toIndex;
                } catch (IOException e) {
                    e.printStackTrace();
                    return null;
                }
            },
            "",
            null/* no initial project */
        );
        //
        //
        //
        //	    while(true) {
        //	    	contentCount++;
        //		    final String content = getContent(contentCount);
        //		    if(content==null) break;
        //	    }

        indexSearchResults(contentToIndex, elURL);

        elasticSearchRepository.waitForIndexingFinished();

        waitForCleanStartUp(contentCount - 1); // TODO this is redundant now to waitForIndexingFinished
    }

    private static final Charset UTF8 = Charset.forName("UTF-8");

    private void indexSearchResults(List<SearchResultDTO> contents, HttpHost elPortAndHost) {
        ObjectMapper mapper = MetaDataRepository.getSearchResultObjectMapper();
        contents.forEach(sr -> {
            try (StringWriter sw = new StringWriter()) {
                mapper.writeValue(sw, sr);
                String contentString = sw.toString();
                RestTemplate restTemplate = new RestTemplate();
                HttpHeaders headers = new HttpHeaders();
                headers.setContentType(MediaType.APPLICATION_JSON);
                HttpEntity<String> request = new HttpEntity<String>(contentString, headers);
                String encodedId = URLEncoder.encode(sr.getExerciseId(), UTF8);
                encodedId = encodedId.replace("/", "%2F");
                URI uri;
                uri = new URI(elPortAndHost.toURI() + "/metadata/_doc/" + encodedId);
                restTemplate.put(uri, request);
            } catch (URISyntaxException | IOException e) {
                LOGGER.warn("Cannot add " + sr.getExerciseId(), e);
            }
        });
    }

    private ExerciseId calculateProjectId(String relativePath) {
        int relativePathPos = relativePath.indexOf("/");
        if (relativePathPos < 0) {
            return new ExerciseId(relativePath, null);
        } else {
            int projectId = Integer.parseInt(relativePath.substring(0, relativePathPos));
            return new ExerciseId(Integer.toString(projectId), relativePath.substring(relativePathPos + 1));
        }
    }

    /**
     * parse the mete data file and return a normalized meta data
     *
     * @param metaDataFile the meta data file
     * @return parsed metadata
     * @throws JsonParseException
     * @throws JsonMappingException
     * @throws IOException
     */
    private UserProvidedMetadataDTO parseMetaDataFile(File metaDataFile) throws JsonParseException, JsonMappingException, IOException {
        ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.findAndRegisterModules();
        UserProvidedMetadataDTO result = mapper.readValue(metaDataFile, UserProvidedMetadataDTO.class);
        makeUpperCase(result.getProgrammingLanguage());
        return result;
    }

    /**
     * make all strings in Array uppercase.
     *
     * @param strings
     */
    private void makeUpperCase(String[] strings) {
        if (strings == null) return;
        for (int i = 0; i < strings.length; i++) {
            strings[i] = strings[i].toUpperCase();
        }
    }

    private GitProjectDTO parseProjectFile(File projectFile) throws JsonParseException, JsonMappingException, IOException {
        ObjectMapper mapper = new ObjectMapper(new YAMLFactory());
        mapper.findAndRegisterModules();
        return mapper.readValue(projectFile, GitProjectDTO.class);
    }

    /**
     * recursively visits all nodes, collects all children metadata and then
     * calls the harvester
     *
     * @param <T>
     *            the type the harvester returns
     * @param dir
     *            the directory to start harvesting in
     * @param harvester
     *            the harvester function
     * @param relativePath
     *            the relative path to dir
     * @param gitProject
     *            the gitProject that may be collected from root.
     * @return the tree of descendants
     */
    private <T> TreeNode<T> visitSubdirectories(File dir, Harvester<T> harvester, String relativePath, GitProjectDTO gitProject) {
        GitProjectDTO currentGitProject = gitProject;
        Assert.state(dir.exists() && dir.isDirectory(), "must be a directory");
        List<TreeNode<T>> children = new ArrayList<TreeNode<T>>();
        for (File containedFile : dir.listFiles()) {
            if (!containedFile.isDirectory() && containedFile.getName().equals("project.yaml")) {
                if (currentGitProject != null) {
                    LOGGER.error(
                        "A project.yaml should not overwrite an existing project in a subnode! {}",
                        containedFile.getAbsolutePath()
                    );
                }
                try {
                    currentGitProject = parseProjectFile(containedFile);
                } catch (IOException e) {
                    LOGGER.error("Cannot parse project.yaml for " + containedFile.getAbsolutePath(), e);
                }
            }
        }
        for (File containedFiles : dir.listFiles()) {
            if (containedFiles.isDirectory()) {
                final TreeNode<T> subResults = visitSubdirectories(
                    containedFiles,
                    harvester,
                    (StringUtils.isEmpty(relativePath) ? "" : (relativePath + "/")) + containedFiles.getName(),
                    currentGitProject
                );
                // special handling of intermediate directories
                if (subResults.getData() != null) {
                    children.add(subResults);
                } else {
                    children.addAll(subResults.getChildren());
                }
            }
        }
        for (File containedFiles : dir.listFiles()) {
            if (!containedFiles.isDirectory() && containedFiles.getName().equals("metaData.yaml")) {
                final T harvest = harvester.harvest(containedFiles, children, relativePath, currentGitProject);
                TreeNode<T> root = new TreeNode<>(harvest);
                root.addChildren(children);
                return root;
            }
        }
        // no metadata found on this level; So we have an empty payload
        TreeNode<T> root = new TreeNode<>(null);
        root.addChildren(children);
        return root;
    }

    /**
     * an interface that is invoked on each node of the tree with a metadata
     *
     * @param <T> the harvested data type
     * @author Michael Breu
     */
    private interface Harvester<T> {
        /**
         * @param f            the metaData File
         * @param subResults   the tree of results of the next node level
         * @param relativePath the relative path from test root
         * @param gitProject   the information about the git project.
         * @return
         */
        T harvest(File f, List<TreeNode<T>> subResults, String relativePath, GitProjectDTO gitProject);
    }

    /**
     * replaces existing metadata by new content
     *
     * @param content
     * @param id
     */
    public void updateMetadata(String content, String id) {
        final RestClientTransport esTransport = (RestClientTransport) elasticsearchAPIClient._transport();

        String elURL = esTransport.restClient().getNodes().get(0).getHost().toURI();
        RestTemplate restTemplate = new RestTemplate();
        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        HttpEntity<String> request = new HttpEntity<String>(content, headers);
        restTemplate.put(elURL + "/metadata/_doc/" + id, request, String.class);
    }

    public String updateMetadata(SearchResultDTO testExercise, String id)
        throws JsonGenerationException, JsonMappingException, IOException {
        final RestClientTransport esTransport = (RestClientTransport) elasticsearchAPIClient._transport();

        String elURL = esTransport.restClient().getNodes().get(0).getHost().toURI();
        // first fetch previous content
        RestTemplate restTemplate = new RestTemplate();
        String previousContent = restTemplate.getForObject(elURL + "/metadata/_doc/" + id, String.class);

        ObjectMapper objectMapper = MetaDataRepository.getSearchResultObjectMapper();
        final JavaTimeModule javaModule = new JavaTimeModule();

        objectMapper.registerModule(javaModule);
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        StringWriter sw = new StringWriter();
        objectMapper.writeValue(sw, testExercise);
        final String contentString = sw.toString();
        updateMetadata(contentString, id);

        return previousContent;
    }

    /**
     * sets the content to a raw json string (containing "_source" attribute). Used
     * to write back a changed value from updateMetadata(...).
     *
     * @param rawInput the json string
     * @param id       the id to write to
     * @throws JsonGenerationException
     * @throws JsonMappingException
     * @throws IOException
     */
    public void resetRawMetadata(String rawInput, String id) throws JsonGenerationException, JsonMappingException, IOException {
        ObjectMapper objectMapper = new ObjectMapper();
        final JavaTimeModule javaModule = new JavaTimeModule();

        objectMapper.registerModule(javaModule);
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        final JsonNode tree = objectMapper.readTree(rawInput);
        final JsonNode source = tree.at("/_source");
        StringWriter sw = new StringWriter();
        objectMapper.writeValue(sw, source);
        final String contentString = sw.toString();
        updateMetadata(contentString, id);
    }

    /**
     * we have to wait until correct startup of content indexing?
     *
     * @param minExpectedHits
     */
    private void waitForCleanStartUp(int minExpectedHits) {
        int tryCount = 0;
        while (tryCount < 10) {
            tryCount++;

            try {
                BoolQuery.Builder b = new BoolQuery.Builder();
                SearchResponse<SearchResultDTO> searchResponse = elasticsearchAPIClient.search(
                    search -> search.index(SearchRepositoryConstants.INDEX_METADATA).query(q -> q.bool(b.build())).from(0).size(1000),
                    SearchResultDTO.class
                );
                @SuppressWarnings("null")
                long totalHits = searchResponse.hits().total().value();
                if (totalHits >= minExpectedHits) {
                    LOGGER.info("Metadata index is up with {} hits after {} trials", totalHits, tryCount);
                    return;
                } else {
                    LOGGER.warn("Metadata index only {} hits up after {} trials", totalHits, tryCount);
                }
            } catch (IOException e) {
                LOGGER.warn("Metadata index startup exception", e);
            }
            try {
                Thread.sleep(1000L);
            } catch (InterruptedException e) {
                LOGGER.debug("Never mind: {}", e.getMessage());
            }
        }
    }

    @EventListener
    public void onApplicationEvent(ApplicationStartingEvent event) {
        try {
            startTestNode();
        } catch (IOException | NodeValidationException e) {
            LOGGER.error("Cannot startup TestNode", e);
        }
    }

    @EventListener
    public void onApplicationEvent(ContextStartedEvent event) {
        try {
            startTestNode();
        } catch (IOException | NodeValidationException e) {
            LOGGER.error("Cannot startup TestNode", e);
        }
    }

    @EventListener
    public void onApplicationEvent(ContextStoppedEvent event) {
        try {
            testNode.close();
        } catch (IOException e) {
            LOGGER.error("Cannot startup TestNode", e);
        }
    }

    public static class TestESNode extends Node {

        public TestESNode(Settings preparedSettings, Collection<Class<? extends Plugin>> classpathPlugins) {
            super(
                InternalSettingsPreparer.prepareEnvironment(preparedSettings, new HashMap<String, String>(), null, null),
                classpathPlugins,
                true
            );
        }
        //		@Override
        //		protected void registerDerivedNodeNameWithLogger(String nodeName) {
        //			LogManager.getLogger(nodeName);
        //		}
    }
}
