package at.ac.uibk.gitsearch.service.gitlab_events;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.verify;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.ElasticSearchRepository;
import at.ac.uibk.gitsearch.repository.search.SearchRepositoryConstants;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.service.MailService;
import at.ac.uibk.gitsearch.service.gitlab_events.GitlabEventService.AnalysedCommit;
import at.ac.uibk.gitsearch.service.gitlab_events.GitlabEventService.Indexing;
import co.elastic.clients.elasticsearch.cat.aliases.AliasesRecord;
import co.elastic.clients.elasticsearch.cat.indices.IndicesRecord;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.exc.StreamReadException;
import com.fasterxml.jackson.databind.DatabindException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectReader;
import com.fasterxml.jackson.databind.PropertyNamingStrategies;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;
import javax.el.MethodNotFoundException;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.elasticsearch.node.NodeValidationException;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.systemhooks.PushSystemHookEvent;
import org.gitlab4j.api.systemhooks.SystemHookEvent;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mockito;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithMockUser;

@SpringBootTest(classes = GitsearchApp.class)
@WithMockUser(value = "test", authorities = "sharing")
public class GitlabEventServiceIT {

    private static final String TEST_META_DATA = "testmetadata";

    static final Logger logger = LoggerFactory.getLogger(GitlabEventServiceIT.class);

    @Autowired
    private GitLabRepository gitLabRepository;

    @Autowired
    protected GitlabEventService eventService;

    @MockBean
    private MailService mailService;

    @Autowired
    protected ElasticSearchRepository elasticSearchRepository;

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the
    // ElasticSearchTestConfiguration
    @Timeout(value = 3, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @Test
    @Timeout(value = 30, unit = TimeUnit.MINUTES)
    void simpleCollectionTest() throws StreamReadException, DatabindException, IOException, GitLabApiException {
        SystemHookEvent collectionCommitEvent = getCollectionCommitEvent();
        Mockito.reset(mailService);
        Mockito
            .when(mailService.sendEmail(Mockito.anyString(), Mockito.anyString(), Mockito.anyString(), Mockito.eq(false), Mockito.eq(true)))
            .thenReturn(Boolean.TRUE);

        logger.info("starting handleGitLabEvent for {}", collectionCommitEvent);
        eventService.handleGitLabEvent(collectionCommitEvent);
        eventService.waitForFinish();
        logger.info("finished handleGitLabEvent for {}", collectionCommitEvent);

        // the example is correct :-)
        verify(mailService, Mockito.times(1))
            .sendEmail(
                Mockito.eq("Michael.Breu@uibk.ac.at"),
                Mockito.anyString(),
                Mockito.anyString(),
                Mockito.eq(false),
                Mockito.eq(true)
            );
    }

    public static Stream<String> allTestSystemHooks() throws IOException {
        return getResourceFiles("testData/gitlabEvents").stream().map(a -> a);
    }

    private static List<String> getResourceFiles(String path) throws IOException {
        List<String> filenames = new ArrayList<>();

        try (
            InputStream in = GitlabEventServiceIT.class.getResourceAsStream(path);
            BufferedReader br = new BufferedReader(new InputStreamReader(in))
        ) {
            String resource;

            while ((resource = br.readLine()) != null) {
                if (resource.endsWith(".json")) {
                    filenames.add(path + "/" + resource);
                } else {
                    // try recursive
                    filenames.addAll(getResourceFiles(path + "/" + resource));
                }
            }
        }

        return filenames;
    }

    @ParameterizedTest
    @MethodSource
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    void allTestSystemHooks(String testResource) throws IOException {
        assertThat(gitLabRepository.getAdminGitLabApi().getAuthToken())
            .describedAs("This test requires a valid admin access token")
            .isNotBlank();
        SystemHookEvent someEvent = getTestSystemHookEvent(testResource);
        assertNotNull(someEvent);
        Mockito.reset(mailService);
        Mockito
            .when(mailService.sendEmail(Mockito.anyString(), Mockito.anyString(), Mockito.anyString(), Mockito.eq(false), Mockito.eq(true)))
            .thenReturn(Boolean.TRUE);

        logger.info("starting handleGitLabEvent for {}", someEvent);
        try {
            eventService.handleGitLabEvent(someEvent);
            eventService.waitForFinish();
        } catch (MethodNotFoundException e) {
            logger.info("Event not supproted {}: {}", someEvent, e.getMessage());
        }
        logger.info("finished handleGitLabEvent for {}", someEvent);
    }

    private SystemHookEvent getCollectionCommitEvent() throws IOException {
        final String testEvent = "testData/gitlabEvents/testInternalCollections.json";
        return getTestSystemHookEvent(testEvent);
    }

    private SystemHookEvent getTestSystemHookEvent(final String testEvent) throws IOException {
        InputStream collectionCommit = this.getClass().getResourceAsStream(testEvent);
        var mapper = new ObjectMapper(new JsonFactory());
        mapper.findAndRegisterModules();
        mapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        mapper.setPropertyNamingStrategy(PropertyNamingStrategies.SNAKE_CASE);
        final ObjectReader jsonReader = mapper.readerFor(SystemHookEvent.class);

        SystemHookEvent collectionCommitEvent = jsonReader.readValue(collectionCommit, SystemHookEvent.class);
        return collectionCommitEvent;
    }

    @Test
    @Timeout(value = 30, unit = TimeUnit.MINUTES)
    void simpleDeleteTest() throws StreamReadException, DatabindException, IOException, GitLabApiException, InterruptedException {
        assertThat(gitLabRepository.getAdminGitLabApi().getAuthToken())
            .describedAs("This test requires a valid admin access token")
            .isNotBlank();
        PushSystemHookEvent collectionCommitEvent = (PushSystemHookEvent) getCollectionCommitEvent();

        final Indexing indexer = eventService.new Indexing(SearchRepositoryConstants.METADATA_PROJECT_ID);

        final Long projectId = collectionCommitEvent.getProjectId();

        final List<String> ids = elasticSearchRepository.getIdsByProjectId(projectId);

        ids.forEach(id -> assertTrue("should start with project id " + projectId, id.startsWith("[" + projectId + "]")));

        final long countBefore = elasticSearchRepository.countByProjectId(projectId);

        assertTrue("Project " + projectId + " should be indexed", countBefore > 0);

        try {
            final long deleteCount = indexer.deleteProjectFromIndex(projectId);
            assertEquals("Project " + projectId + " should be deleted", countBefore, deleteCount);
        } finally {
            // reindex
            Mockito
                .when(
                    mailService.sendEmail(
                        Mockito.anyString(),
                        Mockito.anyString(),
                        Mockito.anyString(),
                        Mockito.eq(false),
                        Mockito.eq(true)
                    )
                )
                .thenReturn(Boolean.TRUE);

            @SuppressWarnings("unused")
            final AnalysedCommit analyzedCommit = eventService.handlePushEvent(collectionCommitEvent);

            elasticSearchRepository.waitForIndexingFinished();
            final long countAfter = elasticSearchRepository.countByProjectId(projectId);
            assertEquals("Project " + projectId + " should be reindexed (May be a race condition)", countBefore, countAfter);

            final List<String> idsByNamespacePrefix = elasticSearchRepository.getIdsByNamespacePrefix(
                collectionCommitEvent.getProject().getNamespace()
            );

            assertTrue(
                "Namespace " + collectionCommitEvent.getProject().getNamespace() + " should contain at least all project ids",
                countAfter <= idsByNamespacePrefix.size()
            );
        }
    }

    @ParameterizedTest
    @ValueSource(strings = { "[3]", "[274]1/b" })
    @Timeout(value = 30, unit = TimeUnit.MINUTES)
    void getDocumentByIdTest(String id) throws GitLabApiException, IOException {
        // should be uploaded as testdata
        final SearchResultDTO searchResultDTO = elasticSearchRepository.get(id);
        assertNotNull(searchResultDTO);
    }

    @ParameterizedTest
    @ValueSource(strings = { "3", "[274]/1/b" })
    @Timeout(value = 30, unit = TimeUnit.MINUTES)
    void testGetDocumentByWrongId(String id) throws GitLabApiException, IOException {
        // should be uploaded as testdata
        final SearchResultDTO searchResultDTO = elasticSearchRepository.get(id);
        assertNull(searchResultDTO);
    }

    @Test
    @Timeout(value = 30, unit = TimeUnit.MINUTES)
    void simpleIndexAllTest() throws GitLabApiException, IOException {
        assertThat(gitLabRepository.getAdminGitLabApi().getAuthToken())
            .describedAs("This test requires a valid admin access token")
            .isNotBlank();

        eventService.indexAllProjects();
    }

    @SuppressWarnings("null")
    @Test
    @Timeout(value = 30, unit = TimeUnit.MINUTES)
    void indexAndAliasManipulationTests() throws GitLabApiException, IOException {
        List<IndicesRecord> indices = elasticSearchRepository.getIndices();
        indices.forEach(ir -> {
            assertNotNull(ir.index());
            assertTrue(ir.index().startsWith("idx_metadata"));
        }); // only metadata indices should exist!

        elasticSearchRepository.createMetadataIndex(TEST_META_DATA);
        indices = elasticSearchRepository.getIndices();
        assertTrue(indices.stream().anyMatch(ir -> ir.index().equals(TEST_META_DATA)));

        for (int i = 0; i < 5; i++) {
            elasticSearchRepository.createMetadataIndex(TEST_META_DATA + i);
        }

        assertEquals(indices.size() + 5, elasticSearchRepository.getIndices().size());

        for (int i = 0; i < 5; i++) {
            elasticSearchRepository.deleteIndex(TEST_META_DATA + i);
        }

        assertEquals(indices.size(), elasticSearchRepository.getIndices().size());

        // test aliases
        List<AliasesRecord> aliases = elasticSearchRepository.getAliases();
        assertEquals(1, aliases.size()); // there should be at least one alias for metadata

        elasticSearchRepository.createAlias(TEST_META_DATA + "Alias", TEST_META_DATA);

        aliases = elasticSearchRepository.getAliases();
        assertEquals(2, aliases.size());

        elasticSearchRepository.deleteAlias(TEST_META_DATA + "Alias", TEST_META_DATA);

        aliases = elasticSearchRepository.getAliases();
        assertEquals(1, aliases.size());
    }
}
