package at.ac.uibk.gitsearch.web.rest;

import static org.assertj.core.api.Assertions.assertThat;
import static org.hamcrest.Matchers.hasItem;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.delete;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.put;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.domain.Statistics;
import at.ac.uibk.gitsearch.repository.jpa.StatisticsRepository;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.StatisticsService;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import at.ac.uibk.gitsearch.service.mapper.StatisticsMapper;
import java.io.IOException;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import javax.persistence.EntityManager;
import org.elasticsearch.node.NodeValidationException;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.junit.jupiter.api.extension.ExtendWith;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.transaction.annotation.Transactional;

/**
 * Integration tests for the {@link StatisticsResource} REST controller.
 */
@SpringBootTest(classes = GitsearchApp.class)
@ExtendWith(MockitoExtension.class)
@AutoConfigureMockMvc
@WithMockUser
class StatisticsResourceIT {

    private static final Integer DEFAULT_VIEWS = 1;
    private static final Integer UPDATED_VIEWS = 2;

    private static final Integer DEFAULT_DOWNLOADS = 1;
    private static final Integer UPDATED_DOWNLOADS = 2;

    private static final String DEFAULT_EXERCISE_ID = "999";

    final String publicGitLabProjectId = "3";

    @Autowired
    private StatisticsRepository statisticsRepository;

    @Autowired
    private StatisticsMapper statisticsMapper;

    @Autowired
    private StatisticsService statisticsService;

    @Autowired
    private EntityManager em;

    @Autowired
    private MockMvc restStatisticsMockMvc;

    private Statistics statistics;

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the ElasticSearchTestConfiguration
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    /**
     * Create an entity for this test.
     *
     * This is a static method, as tests for other entities might also need it, if
     * they test an entity which requires the current entity.
     */
    public static Statistics createEntity(EntityManager em) {
        Statistics statistics = new Statistics().views(DEFAULT_VIEWS).downloads(DEFAULT_DOWNLOADS).exerciseID(DEFAULT_EXERCISE_ID);
        return statistics;
    }

    @BeforeEach
    void initTest() {
        statistics = createEntity(em);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void createStatistics() throws Exception {
        int databaseSizeBeforeCreate = statisticsRepository.findAll().size();
        // Create the Statistics
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(statistics);
        restStatisticsMockMvc
            .perform(
                post("/api/statistics")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(statisticsDTO))
            )
            .andExpect(status().isCreated());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeCreate + 1);
        Statistics testStatistics = statisticsList.get(statisticsList.size() - 1);
        assertThat(testStatistics.getViews()).isEqualTo(DEFAULT_VIEWS);
        assertThat(testStatistics.getDownloads()).isEqualTo(DEFAULT_DOWNLOADS);
        assertThat(testStatistics.getExerciseID()).isEqualTo(DEFAULT_EXERCISE_ID);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void createStatisticsWithExistingId() throws Exception {
        int databaseSizeBeforeCreate = statisticsRepository.findAll().size();

        // Create the Statistics with an existing ID
        statistics.setId(1L);
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(statistics);

        // An entity with an existing ID cannot be created, so this API call must fail
        restStatisticsMockMvc
            .perform(
                post("/api/statistics")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(statisticsDTO))
            )
            .andExpect(status().isBadRequest());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeCreate);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void getAllStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        // Get all the statisticsList
        restStatisticsMockMvc
            .perform(get("/api/statistics?sort=id,desc").with(csrf().asHeader()))
            .andExpect(status().isOk())
            .andExpect(content().contentType(MediaType.APPLICATION_JSON_VALUE))
            .andExpect(jsonPath("$.[*].id").value(hasItem(statistics.getId().intValue())))
            .andExpect(jsonPath("$.[*].views").value(hasItem(DEFAULT_VIEWS)))
            .andExpect(jsonPath("$.[*].downloads").value(hasItem((int) DEFAULT_DOWNLOADS)))
            .andExpect(jsonPath("$.[*].exerciseID").value(hasItem(DEFAULT_EXERCISE_ID)));
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void getStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        // Get the statistics
        restStatisticsMockMvc
            .perform(get("/api/statistics/{id}", statistics.getId()).with(csrf().asHeader()))
            .andExpect(status().isOk())
            .andExpect(content().contentType(MediaType.APPLICATION_JSON_VALUE))
            .andExpect(jsonPath("$.id").value(statistics.getId().intValue()))
            .andExpect(jsonPath("$.views").value(DEFAULT_VIEWS))
            .andExpect(jsonPath("$.downloads").value(DEFAULT_DOWNLOADS))
            .andExpect(jsonPath("$.exerciseID").value(DEFAULT_EXERCISE_ID));
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void testAchievementCalculation() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        // Get the statistics
        restStatisticsMockMvc.perform(get("/user/achievements").with(csrf().asHeader())).andExpect(status().isOk());
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void getNonExistingStatistics() throws Exception {
        // Get the statistics
        restStatisticsMockMvc.perform(get("/api/statistics/{id}", Long.MAX_VALUE).with(csrf().asHeader())).andExpect(status().isNotFound());
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void updateStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        int databaseSizeBeforeUpdate = statisticsRepository.findAll().size();

        // Update the statistics
        Statistics updatedStatistics = statisticsRepository.findById(statistics.getId()).get();
        // Disconnect from session so that the updates on updatedStatistics are not
        // directly saved in db
        em.detach(updatedStatistics);
        updatedStatistics.views(UPDATED_VIEWS).downloads(UPDATED_DOWNLOADS).exerciseID(DEFAULT_EXERCISE_ID);
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(updatedStatistics);

        restStatisticsMockMvc
            .perform(
                put("/api/statistics")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(statisticsDTO))
            )
            .andExpect(status().isOk());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeUpdate);
        Statistics testStatistics = statisticsList.get(statisticsList.size() - 1);
        assertThat(testStatistics.getViews()).isEqualTo(UPDATED_VIEWS);
        assertThat(testStatistics.getDownloads()).isEqualTo(UPDATED_DOWNLOADS);
        assertThat(testStatistics.getExerciseID()).isEqualTo(DEFAULT_EXERCISE_ID);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void updateNonExistingStatistics() throws Exception {
        int databaseSizeBeforeUpdate = statisticsRepository.findAll().size();

        // Create the Statistics
        StatisticsDTO statisticsDTO = statisticsMapper.toDto(statistics);

        // If the entity doesn't have an ID, it will throw BadRequestAlertException
        restStatisticsMockMvc
            .perform(
                put("/api/statistics")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(statisticsDTO))
            )
            .andExpect(status().isBadRequest());

        // Validate the Statistics in the database
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeUpdate);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void deleteStatistics() throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        int databaseSizeBeforeDelete = statisticsRepository.findAll().size();

        // Delete the statistics
        restStatisticsMockMvc
            .perform(delete("/api/statistics/{id}", statistics.getId()).with(csrf().asHeader()).accept(MediaType.APPLICATION_JSON))
            .andExpect(status().isNoContent());

        // Validate the database contains one less item
        List<Statistics> statisticsList = statisticsRepository.findAll();
        assertThat(statisticsList).hasSize(databaseSizeBeforeDelete - 1);
    }

    @Test
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void getStatisticsByExerciseId() throws Exception {
        // Get the statistics and see if it coincides with the mock stats

        statisticsRepository.deleteAll();

        //After the following call a new statistics entity for the exerciseID 143 should have been created because there exists a resource with this id in elastic search from stefan podlipnig
        restStatisticsMockMvc.perform(get("/api/statistics/exercise/{id}", "143").with(csrf().asHeader())).andExpect(status().isOk());
        //wait for the statistics to be created
        //        Thread.sleep(1000);
        assertEquals("143", statisticsService.findOneByExerciseID("143").get().getExerciseID());

        //Do the same call again to see if the statisticsService understood that the entity already exists and it only needs to increase the views
        restStatisticsMockMvc.perform(get("/api/statistics/exercise/{id}", "143").with(csrf().asHeader())).andExpect(status().isOk());

        //wait for the statistics to be created
        //        Thread.sleep(1000);
        assertEquals(2, statisticsService.findOneByExerciseID("143").get().getViews());

        //Test the search for an id which does not exist
        assertEquals(Optional.empty(), statisticsService.findOneByExerciseID("notThere"));
    }

    @ParameterizedTest
    @ValueSource(strings = { "3:with/path" })
    @Transactional
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN)
    void createStatisticsByExerciseIdNotFoundWithES(String exerciseId) throws Exception {
        // Initialize the database
        statisticsRepository.saveAndFlush(statistics);

        restStatisticsMockMvc
            .perform(get("/api/statistics/exercise/{id}", exerciseId).with(csrf().asHeader()))
            .andExpect(status().isNotFound());

        // Get the statistics and see if it is found
        assertEquals(Optional.empty(), statisticsService.findOneByExerciseID(exerciseId));
    }
}
